#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <omp.h>

#define NOT_PRIME 0
#define PRIME 1
#define SET_PRIME 2 

int find_minimum_candidate_prime(unsigned char values[], int value_count, int low_value);
void remove_multiples_gmp(unsigned char values[], int value_count, int low_value, 
                          int global_minimum_prime); 
void display_primes(unsigned char values[], int value_count, int low_value); 

int main(int argc, char *argv[]) {
	const int N = 200000000;
	const int DEBUG = 0; 
	
	unsigned char *values = NULL; 
	int value_count = N, low_value = 0, high_value = N - 1, x, repeat = 1, start; 
	int minimum_candidate_prime, mcp_squared; 
	int prime_count = 0; 
	
	values = (unsigned char *)malloc(sizeof(unsigned char) * value_count); 
	if (values == NULL) {
		printf("malloc() failed\n");
		exit(1); 
	}
		
	for (x = 0; x < value_count; x++)
		values[x] = PRIME; 

	// xxx kludge 0 and 1 on rank 0 are marked out since we start at 2
	values[0] = NOT_PRIME;  
	values[1] = NOT_PRIME; 
	
	while (repeat) {
		minimum_candidate_prime = find_minimum_candidate_prime(values, value_count, low_value); 

		mcp_squared = minimum_candidate_prime * minimum_candidate_prime; 

		if ((mcp_squared >= low_value) && (mcp_squared <= high_value)) {
			start = mcp_squared - low_value; 
			remove_multiples_gmp(values, value_count, start, minimum_candidate_prime); 
		}
			
		//	if ((global_minimum_prime * global_minimum_prime) > high_value)
		//		do nothing;
		
		// Mark the prime, who does it is determined by which range K falls in.
		//
		if ((minimum_candidate_prime >= low_value) && (minimum_candidate_prime <= high_value))
			values[minimum_candidate_prime - low_value] = SET_PRIME; 
		
		if (mcp_squared > N)
			repeat = 0; 
	}

	if (DEBUG) {
		display_primes(values, value_count, low_value); 
		fflush(stdout); 
	}
	
	for (x = 0; x < value_count; x++)
		if (values[x] == SET_PRIME || values[x] == PRIME)
			prime_count++; 

	printf("For N = %d the global_prime_count = %d\n", N, prime_count); 
		
	exit(0); 
}

int find_minimum_candidate_prime(unsigned char values[], int value_count, int low_value) {
	int x = 0;
	int minimum_candidate_prime = 0; 
	
	while ((values[x] == NOT_PRIME || values[x] == SET_PRIME) && (x < value_count)) 
		x++;
		
	if (x < value_count)
		minimum_candidate_prime = x + low_value; 
	else
		minimum_candidate_prime = INT_MAX; 

	return(minimum_candidate_prime); 
}

void remove_multiples_gmp(unsigned char values[], int value_count, int start, 
                          int global_minimum_prime) { 
	int x; 

#pragma omp parallel for num_threads(2) shared(values) private(x)
	for (x = start; x < value_count; x = x + global_minimum_prime)
		values[x] = 0; 

	return;
}

void display_primes(unsigned char values[], int value_count, int low_value) {
	int x = 0;
	
	for (x = 0; x < value_count; x++)
		if (values[x] == SET_PRIME || values[x] == PRIME) {
			printf("x = %d + %d = %d\n", low_value, x, x + low_value); 
			fflush(stdout); 
		}
		
	return; 
}
